/*--------------------------------------------------------------------
 *    The GMT-system:	@(#)pslib.c	2.8  11/27/91
 *
 *    Copyright (c) 1991 by P. Wessel and W. H. F. Smith
 *    See README file for copying and redistribution conditions.
 *--------------------------------------------------------------------*/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * pslib is a library of plot functions that create PostScript.
 * All the routines write their output to the same plotting file,
 * which can be dumped to a Postscript output device (laserwriters).
 * pslib can handle and mix text, line-drawings, and bit-map graphics
 * in both black/white and color.
 *
 * pslib tries to conform to the Encapsulated PostScript Files Specification V 2.0,
 * but makes no claim that including a pslib PostScript file in another
 * application will actually work.  This will hopefully be tested and
 * improved in future releases.
 *
 * C considerations:
 *	All floating point data are assumed to be of type double.
 *	All integer data are assumed to be of type int.
 *	All logical data are assumed to be of type int (1 = TRUE, 0 = FALSE).
 *
 * Updated May, 1992 by J. Goff to include FORTRAN interfaces
 * FORTRAN considerations:
 *	All floating point data are assumed to be DOUBLE PRECISION
 *	All integer data are assumed to be INTEGER, i.e. INTEGER*4
 *	All LOGICAL/int data are assumed to be of type INTEGER*4 (1 = TRUE, 0 = FALSE).
 *
 * List of functions:
 *	ps_arc			: Draws a circular arc
 *	ps_axis			: Plots an axis with tickmarks and annotation/label
 *	ps_circle		: Plots circle and [optionally] fills it
 *	ps_clipoff		: Restores previous clipping path
 *	ps_clipon		: Clips plot outside the specified polygon
 *	ps_colorimage		: Plots a 24-bit 2-D image using the colorimage operator
 *	ps_colortiles		: Plots a 24-bit 2-D image using tiling
 *	ps_comment		: Writes a comment statement to the plot file
 *	ps_cross		: Plots a + 
 *	ps_diamond		: Plots a diamond and [optionally] fills it
 *	ps_ellipse		: Plots an ellipse and [optionally] fills it
 *	ps_image		: Plots a 1-to-8 bit 2-D image using grayshades
 *	ps_imagec		: Plots a 24-bit 2-D image using the NeWS image extension
 *	ps_imagefill		: Fills a polygon with a repeating n x n image pattern
 *	ps_imagergb		: Plots a 24-bit 2-D image using color separations
 *	ps_line			: Plots a line
 *	ps_pie			: Plots a sector of a circle and [optionally] fills it
 *	ps_plot			: Absolute move to new position (pen up or down)
 *	ps_plotend		: Close plotfile
 *	ps_plotinit		: Initialize parameters/open plotfile etc.
 *	ps_plotr		: Relative move to a new position (pen up or down)
 *	ps_polygon		: Creates a polygon and optionally fills it
 *	ps_rect			: Draws a rectangle and [optionally] fills it
 *	ps_rotatetrans		: Rotates, then translates the coordinate system
 *	ps_setdash		: Specify pattern for dashed line
 *	ps_setfont		: Changes current font
 *	ps_setline		: Sets linewidth
 *	ps_setpaint		: Sets the current red/green/blue for fill
 *	ps_square		: Plots square and [optionally] shades it
 *	ps_text			: Plots textstring
 *	ps_textbox		: Draw a filled box around a textstring
 *	ps_transrotate		: Translates and rotates the coordinate system
 *	ps_triangle		: Plots a triangle and [optionally] fills it
 *	ps_vector		: Draws an vector as specified
 *
 *
 * For information about usage, syntax etc, see the pslib.l manual pages
 *
 * Author:	Paul Wessel, Dept. of Geology and Geophysics
 *		School of Ocean and Earth Science and Technology
 *		2525 Correa Road, Honolulu, HI 96822
 *		wessel@soest.hawaii.edu
 * Date:	21-MAY-1991
 * Version:	2.0
 */

#define Version (2.0)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <malloc.h>
#include <memory.h>
#include <time.h>

#include "gmt_parameters.h"
#include "pslib.h"

#define SMALL 1.0e-10
#define MAX_PATH 1200	/* Max PostScript path length */
#define R2D (180.0/M_PI)
#define I_255 (1.0 / 255.0)

/* Misc macros */

#define rint(x) ((double)(int)(((x) < 0.0) ? (x) - 0.5 : (x) + 0.5))
#define ceil(x) ((double)(int)(((x) < 0.0 || (x) == (int)(x)) ? (x) : (x) + 1.0))
#define floor(x) ((double)(int)(((x) > 0.0 || (x) == (int)(x)) ? (x) : (x) - 1.0))

int ps_arc (x, y, radius, az1, az2, status)
double x, y, radius, az1, az2;
int status; {	/* 1 = set anchor, 2 = set end, 3 = both */
	int ix, iy, ir;
	ix = rint (x * ps_scl);
	iy = rint (y * ps_scl);
	ir = rint (radius * ps_scl);
	if (fabs (az1 - az2) > 360.0) az1 = 0.0, az2 = 360.0; 
	if (status%2) {	/* Beginning of new segment */
		fprintf (ps_file, "S ");
		ps_npath = 0;
	}
	else
		ps_npath++;
	if (az1 < az2)	/* Forward positive arc */
		fprintf (ps_file, "%d %d %d %lg %lg arc", ix ,iy, ir, az1, az2);
	else	/* Negative arc */
		fprintf (ps_file, "%d %d %d %lg %lg arcn", ix ,iy, ir, az1, az2);
	if (status > 1)	fprintf (ps_file, " S");
	fprintf (ps_file, "\n");
}

/* fortran interface */
int ps_arc_ (x, y, radius, az1, az2, status)
double *x, *y, *radius, *az1, *az2;
int *status; {
	 ps_arc (*x, *y, *radius, *az1, *az2, *status);
}
	
int ps_axis (x, y, length, val0, val1, anotation_int, label, anotpointsize, side)
double x, y, length, val0, val1, anotation_int;
int anotpointsize, side;
char *label; {
	int anot_justify, label_justify, i, j, ndig = 0;
	int left = FALSE;
	double angle, dy, scl, val, anot_off, label_off, xx, sign;
	char text[80], format[20];
	
	if (anotation_int < 0.0) left = TRUE;
	anotation_int = fabs (anotation_int);
	sprintf (text, "%lg\0", anotation_int);
	for (i = 0; text[i] && text[i] != '.'; i++);
	if (text[i]) {	/* Found a decimal point */
		for (j = i + 1; text[j]; j++);
		ndig = j - i - 1;
	}
	if (ndig > 0)
		sprintf (format, "%%.%dlf\0", ndig);
	else
		strcpy (format, "%lg");
		
	angle = (side%2) ? 90.0 : 0.0;
	sign = (side < 2) ? -1.0 : 1.0;
	anot_justify = label_justify = (side < 2) ? -10 : -2;
	dy = sign * anotpointsize / 72.0;
			
	fprintf (ps_file, "\nV %d %d T %lg R\n", (int)rint (x * ps_scl), (int)rint (y * ps_scl), angle);
	ps_plot (0.0, 0.0, 3);
	ps_plot (length, 0.0, 2);
	if ((val1 - val0) == 0.0) {
		fprintf (stderr, "pslib: ERROR: Axis val0 == val1!\n");
		return;
	}
	scl = length / (val1 - val0);
	anot_off = dy;
	label_off = 2.5 * dy;
	dy *= 0.5;
	
	i = 0;
	val = val0;
	while (val <= (val1+SMALL)) {
		i++;
		xx = (val - val0) * scl;
		if (left) xx = length - xx;
		ps_plot (xx, 0.0, 3);
		ps_plot (xx, dy, 2);
		sprintf( text, format, val);
		ps_text (xx, anot_off, anotpointsize, text, 0.0, anot_justify, 0);
		val = val0 + i * anotation_int;
	}
	ps_text (0.5*length, label_off, (int) (anotpointsize*1.5), label, 0.0, label_justify, 0);
	fprintf (ps_file, "U\n\n");
}

/* fortran interface */
int ps_axis_ (x, y, length, val0, val1, anotation_int, label, anotpointsize, side, nlen)
double *x, *y, *length, *val0, *val1, *anotation_int;
int *anotpointsize, *side;
int nlen;
char *label; {
	ps_axis (*x, *y, *length, *val0, *val1, *anotation_int, label, *anotpointsize, *side);
}

int ps_circle (x, y, size, r, g, b, outline)
double x, y, size;
int r, g, b;
int outline; {
	int ix, iy, ir;
	double rr, gg, bb;
	
	rr = r * I_255;
	ix = rint (x * ps_scl);
	iy = rint (y * ps_scl);
	ir = rint (size * ps_scl);
	if (r == g && g == b && b == g)	/* Grayshade only */
		fprintf (ps_file, "N %.3lg %d %d %d C%d\n", rr, ix, iy, ir, outline);
	else {	/* Color */
		gg = g * I_255;	bb = b * I_255;
		fprintf (ps_file, "N %.3lg %.3lg %.3lg %d %d %d C%d\n", rr, gg, bb, ix, iy, ir, outline + 2);
	}
	ps_npath = 0;
}

/* fortran interface */
int ps_circle_ (x, y, size, r, g, b, outline)
double *x, *y, *size;
int *r, *g, *b;
int *outline; {
	 ps_circle (*x, *y, *size, *r, *g, *b, *outline);
}

int ps_clipoff () {
	fprintf (ps_file, "\nS U\n");
	ps_comment ("Clipping is currently OFF\n");
	ps_npath = 0;
}

/* fortran interface */
int ps_clipoff_ () {
	ps_clipoff ();
}

int ps_clipon (x, y, n, r, g, b, flag)
double x[], y[];
int n, r, g, b, flag; {
	/* Any plotting outside the path defined by x,y will be clipped.
	use clipoff to restore the original clipping path. */
	
	int used;
	double rr, gg, bb;
	char move[7];
	
	rr = r * I_255;
	
	if (flag%2) ps_npath = 0;
	if (flag%2) {	/* First segment in multi-segment clip-path */
		strcpy (move, "M");
		fprintf (ps_file, "\n");
		ps_comment ("Start of clip path");
		fprintf (ps_file, "S V\n");
		ps_npath = 0;
	}
	else
		strcpy (move, "moveto");
		
	used = ps_npath;
	if (n > 0) {
		ps_ix = rint (x[0]*ps_scl);
		ps_iy = rint (y[0]*ps_scl);
		ps_npath++;
		used++;
		fprintf (ps_file, "%d %d %s\n", ps_ix, ps_iy, move);
		used += ps_line (&x[1], &y[1], n-1, 0, TRUE);
	}
	if (used > MAX_PATH) fprintf (stderr, "pslib: clipon called with > MAX_PATH points (%d)!\n", used);
	if (flag >= 2) {
		if (r >= 0 && (r == g && g == b && b == g))
			fprintf (ps_file, "V %.3lg A eofill U ", rr);
		else if (r >= 0) {
			gg = g * I_255;	bb = b * I_255;
			fprintf (ps_file, "V %.3lg %.3lg %.3lg C eofill U ", rr, gg, bb);
		}
		fprintf (ps_file, "eoclip N\n");
		ps_comment ("End of clip path.  Clipping is currently ON\n");
		ps_npath = 0;
	}
}

/* fortran interface */
int ps_clipon_ (x, y, n, r, g, b, flag)
double x[], y[];
int *n, *r, *g, *b, *flag; {
	ps_clipon (x, y, *n, *r, *g, *b, *flag);
}
	  
int ps_colorimage (x, y, xsize, ysize, buffer, nx, ny)
double x, y, xsize, ysize;
unsigned char *buffer;
int nx, ny; {
	/* Plots a color bitmapped image. Each pixel is described by
	 * 3 hexadecimal numbers, each may take on values from
	 * 00(0) to FF(255).  buffer stores the colroinfo in
	 * r g b r g b r g b format.
	 */
	char hex[16], pixel[61];
	int ix, iy, lx, ly, mx, i, j, kk, ij;
	
	pixel[60] = 0;
	hex[0] = '0';	hex[1] = '1';	hex[2] = '2';	hex[3] = '3';
	hex[4] = '4';	hex[5] = '5';	hex[6] = '6';	hex[7] = '7';
	hex[8] = '8';	hex[9] = '9';	hex[10] = 'A';	hex[11] = 'B';
	hex[12] = 'C';	hex[13] = 'D';	hex[14] = 'E';	hex[15] = 'F';
	
	ix = rint (x * ps_scl);
	iy = rint (y * ps_scl);
	lx = rint (xsize * ps_scl);
	ly = rint (ysize * ps_scl);
	mx = 3 * nx;
	fprintf (ps_file, "\n%% Start of colorimage\n");
	fprintf (ps_file, "V N %d %d T %d %d scale\n", ix, iy, lx, ly);
	fprintf (ps_file, "%d %d 8 [%d 0 0 %d 0 %d] {currentfile %d string readhexstring pop} false 3 colorimage\n",
		nx, ny, nx, -ny, ny, mx);
	kk = 0;
	for (j = ij = 0; j < ny; j++) {
		for (i = 0; i < nx; i++) {
			pixel[kk++] = hex[buffer[ij] / 16];
			pixel[kk++] = hex[buffer[ij++] % 16];
			pixel[kk++] = hex[buffer[ij] / 16];
			pixel[kk++] = hex[buffer[ij++] % 16];
			pixel[kk++] = hex[buffer[ij] / 16];
			pixel[kk++] = hex[buffer[ij++] % 16];
			if (kk == 60) {
				fprintf (ps_file, "%s\n", pixel);
				kk = 0;
			}
		}
	}
	if (kk > 0) {
		pixel[kk] = 0;
		fprintf (ps_file, "%s\n", pixel);
	}
	fprintf (ps_file, "U\n%% End of colorimage\n\n");
}

/* fortran interface */
int ps_colorimage_ (x, y, xsize, ysize, buffer, nx, ny, nlen)
double *x, *y, *xsize, *ysize;
unsigned char *buffer;
int nlen;
int *nx, *ny; {
	ps_colorimage (*x, *y, *xsize, *ysize, buffer, *nx, *ny);
}

int ps_colortiles (x0, y0, dx, dy, image, nx, ny)
double x0, y0;		/* Lower left corner in inches */
double dx, dy;		/* Size of cell in inches */
unsigned char *image;	/* color image  */
int nx, ny; {		/* image size */
	int i, j, k, r, g, b;
	double x1, x2, y1, y2, noise;
	
	noise = 2.0 / ps_scl;
	dx /= nx;
	dy /= ny;

	ps_transrotate (x0, y0, 0.0);
	y2 = (ny - 0.5) * dy + 0.5 * noise;
	for (j = k = 0; j < ny; j++) {
		y1 = (ny - j - 1.5) * dy - 0.5 * noise;
		x1 = -0.5 * (dx + noise);
		for (i = 0; i < nx; i++) {
			x2 = (i + 0.5) * dx + noise;
			r = image[k++];
			g = image[k++];
			b = image[k++];
			ps_rect (x1, y1, x2, y2, r, g, b, FALSE);
			x1 = x2 - 2.0*noise;
		}
		y2 = y1 + 2.0*noise;
	}
	ps_rotatetrans (-x0, -y0, 0.0);
}

/* fortran interface */
int ps_colortiles_ (x0, y0, dx, dy, image, nx, ny, nlen)
double *x0, *y0;
double *dx, *dy;
unsigned char *image;
int nlen;
int *nx, *ny; {
	 ps_colortiles (*x0, *y0, *dx, *dy, image, *nx, *ny);
}

int ps_comment (text)
char *text; {
	fprintf (ps_file, "%% %s\n", text);
}

/* fortran interface */
int ps_comment_ (text, nlen)
int nlen;
char *text; {
	ps_comment (text);
}

int ps_cross (x, y, size)
double x, y, size; {
	int ix ,iy, dx;
	
	ix = rint ((x - 0.5 * size) * ps_scl);
	iy = rint (y * ps_scl);
	dx = rint (size * ps_scl);
	fprintf (ps_file, "%d %d %d X\n", dx, ix, iy);
	ps_npath = 0;
}

/* fortran interface */
int ps_cross_ (x, y, size)
double *x, *y, *size; {
	 ps_cross (*x, *y, *size);
}

int ps_diamond (x, y, side, r, g, b, outline)
double x, y, side;
int r, g, b;
int outline; {	/* see circle */
	int ix, iy, ds;
	double rr, gg, bb;
	
	side /= M_SQRT2;
	rr = r * I_255;
	ds = rint (side * ps_scl);
	ix = rint (x * ps_scl);
	iy = rint ((y - side) * ps_scl);
	if (r == g && g == b && b == g) /* Grayshade only */
		fprintf (ps_file, "%.3lg %d %d %d D%d\n", rr, ds, ix, iy, outline);
	else {	/* Color */
		gg = g * I_255;	bb = b * I_255;
		fprintf (ps_file, "%.3lg %.3lg %.3lg %d %d %d D%d\n", rr, gg, bb, ds, ix, iy, outline+2);
	}
	ps_npath = 0;
}

/* fortran interface */
int ps_diamond_ (x, y, side, r, g, b, outline)
double *x, *y, *side;
int *r, *g, *b;
int *outline; {
	 ps_diamond (*x, *y, *side, *r, *g, *b, *outline);
}

int ps_ellipse (x, y, angle, major, minor, r, g, b, outline)
double x, y, angle, major, minor;
int r, g, b, outline; {
	int ir;
	double aspect, rr, gg, bb;
	
	fprintf (ps_file, "V %d %d T", rint (x * ps_scl), rint (y * ps_scl));
	if (angle != 0.0) fprintf (ps_file, " %lg R", angle);
	aspect = minor / major;
	fprintf (ps_file, " 1 %lg scale\n", aspect);
	rr = r * I_255;
	ir = rint (major * ps_scl);
	if (r == g && g == b && b == g)	/* Grayshade only */
		fprintf (ps_file, " %.3lg 0 0 %d C%d U\n", rr, ir, outline);
	else {	/* Color */
		gg = g * I_255;	bb = b * I_255;
		fprintf (ps_file, " %.3lg %.3lg %.3lg 0 0 %d C%d U\n", rr, gg, bb, ir, outline + 2);
	}
}

/* fortran interface */
int ps_ellipse_ (x, y, angle, major, minor, r, g, b, outline)
double *x, *y, *angle, *major, *minor;
int *r, *g, *b, *outline; {
	 ps_ellipse (*x, *y, *angle, *major, *minor, *r, *g, *b, *outline);
}

int ps_image (x, y, xsize, ysize, buffer, nx, ny, nbits)
double x, y, xsize, ysize;
unsigned char *buffer;
int nx, ny, nbits; {
	/* Plots a bitmapped image. Each pixel is described by
	 * a hexadecimal number, which may take on values from
	 * 00(0) to FF(255).
	 * nbits is no of bits/pixel. Can be 8,4, or 1.
	 * Eg. if 1 is used, the each element of buffer provides
	 * shade info for 8 pixels, going from left to right.
	 * Note that nx refers to number of buffer values, not pixels.
	 * nx must be an integral of 8/nbits.
	 */
	char hex[16], pixel[80];
	int ix, iy, lx, ly, mx, i, j, kk, jj;
	
	hex[0] = '0';	hex[1] = '1';	hex[2] = '2';	hex[3] = '3';
	hex[4] = '4';	hex[5] = '5';	hex[6] = '6';	hex[7] = '7';
	hex[8] = '8';	hex[9] = '9';	hex[10] = 'A';	hex[11] = 'B';
	hex[12] = 'C';	hex[13] = 'D';	hex[14] = 'E';	hex[15] = 'F';
	
	ix = rint (x * ps_scl);
	iy = rint (y * ps_scl);
	lx = rint (xsize * ps_scl);
	ly = rint (ysize * ps_scl);
	mx = nx * 8 / nbits;
	fprintf (ps_file, "\n%% Start of image\n");
	fprintf (ps_file, "V N %d %d T %d %d scale\n", ix, iy, lx, ly);
	fprintf (ps_file, "%d %d %d [%d 0 0 %d 0 %d] {currentfile %d string readhexstring pop} image\n",
		mx, ny, nbits, mx, -ny, ny, nx);
	kk = 0;
	for (j = jj = 0; j < ny; j++) {
		for (i = 0; i < nx; i++, jj++) {
			pixel[kk++] = hex[buffer[jj]/16];
			pixel[kk++] = hex[buffer[jj]%16];
			if (kk == 80) {
				for (kk = 0; kk < 80; kk++) putc (pixel[kk], ps_file);
				putc ('\n', ps_file);
				kk = 0;
			}
		}
	}
	if (kk > 0) {
		for (i = 0; i < kk; i++) putc (pixel[i], ps_file);
		putc ('\n', ps_file);
	}
	fprintf (ps_file, "U\n%% End of image\n\n");
}

/* fortran interface */
int ps_image_ (x, y, xsize, ysize, buffer, nx, ny, nbits, nlen)
double *x, *y, *xsize, *ysize;
unsigned char *buffer;
int nlen;
int *nx, *ny, *nbits; {
	ps_image (*x, *y, *xsize, *ysize, buffer, *nx, *ny, *nbits);
}

int ps_imagec (x, y, xsize, ysize, buffer, nx, ny)
double x, y, xsize, ysize;
unsigned char *buffer;
int nx, ny; {
	/* Plots a color bitmapped image. Each pixel is described by
	 * 3 hexadecimal numbers, each may take on values from
	 * 00(0) to FF(255).  buffer stores the colroinfo in
	 * r g b r g b r g b format.
	 * Temporary version untill colorimage is defined
	 */
	char hex[16], pixel[61];
	int ix, iy, lx, ly, mx, i, j, kk, ij;
	
	pixel[60] = 0;
	hex[0] = '0';	hex[1] = '1';	hex[2] = '2';	hex[3] = '3';
	hex[4] = '4';	hex[5] = '5';	hex[6] = '6';	hex[7] = '7';
	hex[8] = '8';	hex[9] = '9';	hex[10] = 'A';	hex[11] = 'B';
	hex[12] = 'C';	hex[13] = 'D';	hex[14] = 'E';	hex[15] = 'F';
	
	ix = rint (x * ps_scl);
	iy = rint (y * ps_scl);
	lx = rint (xsize * ps_scl);
	ly = rint (ysize * ps_scl);
	mx = 3 * nx;
	fprintf (ps_file, "\n%% Start of colorimage\n");
	fprintf (ps_file, "V N %d %d T %d %d scale\n", ix, iy, lx, ly);
	fprintf (ps_file, "%d %d 24 [%d 0 0 %d 0 %d] {currentfile %d string readhexstring pop} image\n",
		nx, ny, nx, -ny, ny, mx);
	kk = 0;
	for (j = ij = 0; j < ny; j++) {
		for (i = 0; i < nx; i++) {
			pixel[kk++] = hex[buffer[ij] / 16];
			pixel[kk++] = hex[buffer[ij++] % 16];
			pixel[kk++] = hex[buffer[ij] / 16];
			pixel[kk++] = hex[buffer[ij++] % 16];
			pixel[kk++] = hex[buffer[ij] / 16];
			pixel[kk++] = hex[buffer[ij++] % 16];
			if (kk == 60) {
				fprintf (ps_file, "%s\n", pixel);
				kk = 0;
			}
		}
	}
	if (kk > 0) {
		pixel[kk] = 0;
		fprintf (ps_file, "%s\n", pixel);
	}
	fprintf (ps_file, "U\n%% End of colorimage\n\n");
}

/* fortran interface */
int ps_imagec_ (x, y, xsize, ysize, buffer, nx, ny, nlen)
double *x, *y, *xsize, *ysize;
unsigned char *buffer;
int nlen;
int *nx, *ny; {
	 ps_imagec (*x, *y, *xsize, *ysize, buffer, *nx, *ny);
}

int ps_imagefill (x, y, n, image_no, imagefile, invert, image_size, outline)
double x[], y[];
double image_size;	/* Size of image on page in inches */
int n;			/* No of points in path */
int image_no;		/* no of image (1-32) or user (0) */
int invert;		/* If invert == TRUE we interchange black and white pixels */
char *imagefile;	/* Name of image file if image == 0 */
int outline; {
	int i, j, nx, ny, ix, iy, dx, dump = FALSE, n_times = 0;
	char *string, tmp[65], *get_icon (), name[10], file[512];
	double xx, yy, xmin, xmax, ymin, ymax;
	
	if (image_size <= 0.0) {	/* Use device resolution */
		dx = 64;
		image_size = 64.0 / ps_scl;
	}
	else
		dx = rint (image_size * ps_scl);
	
	image_no--;	/* Frome 1-32 to 0-31 */
	if (image_no >= 0 && image_no < N_PATTERNS && ps_pattern_status[image_no][invert] == 0) {
		sprintf (file, "%s/ps_pattern.%d\0", GMTLIBDIR, image_no);
		string = get_icon (file, &nx, &ny, invert);
	}
	else if (image_no < 0)
		string = get_icon (imagefile, &nx, &ny, invert);
	else if (ps_pattern_status[image_no][invert] == 1) {
		nx = ps_pattern_nx[image_no][invert];
		ny = ps_pattern_ny[image_no][invert];
		dx = ps_pattern_dx[image_no][invert];
	}
	ps_comment ("\n%Start of user imagefill pattern");
	if (image_no < 0) {
		strcpy (name, "userimage");
		fprintf (ps_file, "/userimage <\n");
		dump = TRUE;
	}
	else {
		if (invert)
			sprintf (name, "image%di\0", image_no);
		else
			sprintf (name, "image%d\0", image_no);
		if (ps_pattern_status[image_no][invert] == 0) {
			fprintf (ps_file, "/%s <\n", name);
			ps_pattern_status[image_no][invert] = 1;
			ps_pattern_nx[image_no][invert] = nx;
			ps_pattern_ny[image_no][invert] = ny;
			ps_pattern_dx[image_no][invert] = dx;
			dump = TRUE;
		}
	}
	if (dump) {
		for (i = 0; i < nx * ny / 256; i++) {
			strncpy (tmp, &string[i*64], 64);
			tmp[64] = 0;
			fprintf (ps_file, "\t%s\n", tmp);
		}
		fprintf (ps_file, "> def\n");
	}
	fprintf (ps_file, "/imagefill { V T %d dup scale %d %d 1 [%d 0 0 %d 0 %d] {%s} image U} def\n", dx, nx, ny, nx, -ny, ny, name);
	if (dump) free (string);
	
	/* Print out clip-path */
	
	ps_clipon (x, y, n, -1, -1, -1, 3);
	
	/* Find extreme bounds for area */
	
	xmin = xmax = x[0];
	ymin = ymax = y[0];
	for (i = 1; i < n; i++) {
		xmin = MIN (xmin, x[i]);
		ymin = MIN (ymin, y[i]);
		xmax = MAX (xmax, x[i]);
		ymax = MAX (ymax, y[i]);
	}
	
	for (j = (int) floor (ymin / image_size); j <= (int) ceil (ymax / image_size); j++) {
		yy = j * image_size;
		for (i = (int) floor (xmin / image_size); i <= (int) ceil (xmax / image_size); i++) {
			xx = i * image_size;
			ix = rint (xx * ps_scl);
			iy = rint (yy * ps_scl);
			fprintf (ps_file, "%d %d", ix, iy);
			n_times++;
			(n_times%5) ? putc (' ', ps_file) : putc ('\n', ps_file);
		}
	}
	if (n_times%5) putc ('\n', ps_file);
	fprintf (ps_file, "%d {imagefill} repeat\n", n_times);
	if (outline) fprintf (ps_file, "clippath S\n");
	ps_clipoff ();
	ps_comment ("End of user imagefill pattern\n");
}

/* fortan interface */
int ps_imagefill_ (x, y, n, image_no, imagefile, invert, image_size, outline, nlen)
double x[], y[];
double *image_size; 
int *n;
int *image_no; 
int *invert; 
char *imagefile;
int nlen;
int *outline; {
	 ps_imagefill (x, y, *n, *image_no, imagefile, *invert, *image_size, *outline);
}

int ps_imagergb (x, y, xsize, ysize, buffer, nx, ny)
double x, y, xsize, ysize;
unsigned char *buffer;
int nx, ny; {
	/* Plots a 24bit bitmapped image using three calls to image 
	   called rimage, gimage, bimage.  These will be defined and
	   undefined when psto24 is run.
	 * Temporary version untill colorimage is defined
	 */
	
	char hex[16], pixel[80], comp[3];
	int ix, iy, lx, ly, i, j, kk, jj, color;
	
	hex[0] = '0';	hex[1] = '1';	hex[2] = '2';	hex[3] = '3';
	hex[4] = '4';	hex[5] = '5';	hex[6] = '6';	hex[7] = '7';
	hex[8] = '8';	hex[9] = '9';	hex[10] = 'A';	hex[11] = 'B';
	hex[12] = 'C';	hex[13] = 'D';	hex[14] = 'E';	hex[15] = 'F';
	comp[0] = 'r';	comp[1] = 'g';	comp[2] = 'b';
	ix = rint (x * ps_scl);
	iy = rint (y * ps_scl);
	lx = rint (xsize * ps_scl);
	ly = rint (ysize * ps_scl);
	fprintf (ps_file, "\n%% Start of colorimage\n");
	fprintf (ps_file, "V N %d %d T %d %d scale\n", ix, iy, lx, ly);
	
	for (color = jj = 0; color < 3; color++) {
		fprintf (ps_file, "psto24pass %d eq\n", color);
		fprintf (ps_file, "{%d %d 8 [%d 0 0 %d 0 %d] {currentfile %d string readhexstring pop} image}\n",
			nx, ny, nx, -ny, ny, nx);
		fprintf (ps_file, "{%d {currentfile %d string readhexstring pop pop} repeat}\n", ny, nx);
		fprintf (ps_file, "ifelse\n");
		kk = 0;
		for (j = 0; j < ny; j++) {
			for (i = 0; i < nx; i++, jj++) {
				pixel[kk++] = hex[buffer[jj]/16];
				pixel[kk++] = hex[buffer[jj]%16];
				if (kk == 80) {
					for (kk = 0; kk < 80; kk++) putc (pixel[kk], ps_file);
					putc ('\n', ps_file);
					kk = 0;
				}
			}
		}
		if (kk > 0) {
			for (i = 0; i < kk; i++) putc (pixel[i], ps_file);
			putc ('\n', ps_file);
		}
	}
	fprintf (ps_file, "U\n%% End of imagergb\n\n");
}

/* fortran interface */
int ps_imagergb_ (x, y, xsize, ysize, buffer, nx, ny, nlen)
double *x, *y, *xsize, *ysize;
unsigned char *buffer;
int nlen;
int *nx, *ny; {
	ps_imagergb (*x, *y, *xsize, *ysize, buffer, *nx, *ny);
}

int ps_line (x, y, n, type, close)
double *x, *y;
int n, type;	/* type: 1 means new anchor point, 2 means stroke line, 3 = both */
int close; {
	int i, *ix, *iy;
	
	/* First remove unneccessary points that have zero curvature */
	
	ix = (int *) malloc ((unsigned) (n * sizeof (int)));
	iy = (int *) malloc ((unsigned) (n * sizeof (int)));
	
	n = ps_shorten_path (x, y, n, ix, iy);
	if (type%2) {
		fprintf (ps_file, "%d %d M\n", ix[0], iy[0]);
		ps_npath = 1;
	}
	else
		fprintf (ps_file, "%d %d D\n", ix[0] - ps_ix, iy[0] - ps_iy);
	ps_ix = ix[0];
	ps_iy = iy[0];
		
	for (i = 1; i < n; i++) {
		fprintf (ps_file, "%d %d D\n", ix[i] - ps_ix, iy[i] - ps_iy);
		ps_ix = ix[i];
		ps_iy = iy[i];
		ps_npath++;
		if (ps_npath > MAX_PATH) {
			fprintf (ps_file, "S %d %d M\n", ps_ix, ps_iy);
			ps_npath = 1;
		}
	}
	if (close) fprintf (ps_file, "P");	/* Close the path */
	if (type > 1) {
		fprintf (ps_file, " S\n");	/* Stroke the path */
		ps_npath = 0;
	}
	else
		fprintf (ps_file, "\n");
	
	free ((char *)ix);
	free ((char *)iy);
	
	return (n);
}

/* fortran interface */
int ps_line_ (x, y, n, type, close)
double x[], y[];
int *n, *type;
int *close; {
	ps_line (x, y, *n, *type, *close);
}

int ps_shorten_path (x, y, n, ix, iy)
double x[], y[];
int n;
int ix[], iy[]; {
	double old_slope, new_slope, dx, dy;
	int i, k, *xx, *yy, fixed;
	
	xx = (int *) malloc ( (unsigned) (n * sizeof (int)));
	yy = (int *) malloc ( (unsigned) (n * sizeof (int)));
	
	for (i = 0; i < n; i++) {
		xx[i] = rint (x[i] * ps_scl);
		yy[i] = rint (y[i] * ps_scl);
	}
	
	ix[0] = xx[0];	iy[0] = yy[0];	k = 1;
	
	dx = xx[1] - xx[0];
	dy = yy[1] - yy[0];
	fixed = (dx == 0.0 && dy == 0.0);
	old_slope = (fixed) ? 1.01e100 : ((dx == 0) ? copysign (1.0e100, dy) : dy / dx);
	
	for (i = 1; i < n-1; i++) {
		dx = xx[i+1] - xx[i];
		dy = yy[i+1] - yy[i];
		fixed = (dx == 0.0 && dy == 0.0);
		new_slope = (fixed) ? 1.01e100 : ((dx == 0) ? copysign (1.0e100, dy) : dy / dx);
		if (fixed) continue;	/* Didnt move */
		
		if (new_slope != old_slope) {
			ix[k] = xx[i];
			iy[k] = yy[i];
			k++;
			old_slope = new_slope;
		}
	}
	dx = xx[n-1] - xx[n-2];
	dy = yy[n-1] - yy[n-2];
	fixed = (dx == 0.0 && dy == 0.0);	/* Didnt move */
	if (!fixed) {
		ix[k] = xx[n-1];
		iy[k] = yy[n-1];
		k++;
	}
	
	free ((char *)xx);
	free ((char *)yy);
	
	return (k);
}

/* fortran interface */
int ps_shorten_path_ (x, y, n, ix, iy)
double x[], y[];
int *n;
int ix[], iy[]; {
	ps_shorten_path (x, y, *n, ix, iy);
}


int ps_line_old (x, y, n, type, close)
double *x, *y;
int n, type;	/* type: 1 means new anchor point, 2 means stroke line, 3 = both */
int close; {
	int i;
	
	ps_plot (x[0], y[0], 2 + type%2);	/* Gives 3 if type = 1 || 3 */
	for (i = 1; i < n; i++) ps_plot (x[i], y[i], 2);
	if (close) fprintf (ps_file, "P");	/* Close the path */
	ps_npath += n;
	if (type > 1) {
		fprintf (ps_file, " S\n");	/* Stroke the path */
		ps_npath = 0;
	}
	else
		fprintf (ps_file, "\n");
}

int ps_pie (x, y, radius, az1, az2, r, g, b, outline)
double x, y, radius, az1, az2;
int r, g, b;
int outline; {
	int ix, iy, ir;
	double rr, gg, bb;
	
	rr = r * I_255;
	ix = rint (x * ps_scl);
	iy = rint (y * ps_scl);
	ir = rint (radius * ps_scl);
	if (r == g && g == b && b == g) /* Grayshade */
		fprintf (ps_file, "%d %d M %.3lg %d %d %d %lg %lg P%d\n", ix, iy, rr, ix, iy, ir, az1, az2, outline);
	else {
		gg = g * I_255;	bb = b * I_255;
		fprintf (ps_file, "%d %d M %.3lg %.3lg %.3lg %d %d %d %lg %lg P%d\n",
			ix, iy, rr, gg, bb, ix, iy, ir, az1, az2, outline+2);
	}
	ps_npath = 0;
}

/* fortran interface */
int ps_pie_ (x, y, radius, az1, az2, r, g, b, outline)
double *x, *y, *radius, *az1, *az2;
int *r, *g, *b;
int *outline; {
	 ps_pie (*x, *y, *radius, *az1, *az2, *r, *g, *b, *outline);
}

int ps_plot (x, y, pen)
double x, y;
int pen; {
	int ix, iy, idx, idy;
	ix = rint (x*ps_scl);
	iy = rint (y*ps_scl);
	if (pen == 2) {	/* Convert absolute draw to relative draw */
		idx = ix - ps_ix;
		idy = iy - ps_iy;
	}
	else {
		idx = ix;
		idy = iy;
	}
	if (pen == 2 && idx == 0 && idy == 0) return;
	ps_npath++;
	ps_ix = ix;
	ps_iy = iy;
	if (pen == 2)
		fprintf (ps_file, "%d %d D\n", idx, idy);
	else {
		fprintf (ps_file, "%d %d M\n", idx, idy);
		ps_npath = 1;
	}
	if (ps_npath > MAX_PATH) {
		fprintf (ps_file, "S %d %d M\n", ix, iy);
		ps_npath = 1;
	}
}

/* fortran interface */
int ps_plot_ (x, y, pen)
double *x, *y;
int *pen; {
	ps_plot (*x, *y, *pen);
}

int ps_plotend (lastpage)
int lastpage; {
	fprintf (ps_file, "S");
	if (lastpage) {
		fprintf (ps_file, " %d %d T",-(int)rint(ps.xoff*ps_scl),-(int)rint(ps.yoff*ps_scl));
		fprintf (ps_file, " %lg %lg scale",ps_scl/(72.0 * ps.xscl), ps_scl/(72.0 * ps.yscl));
		if (ps.mode = 0) fprintf (ps_file, " -90 R -612 0 T");
		fprintf (ps_file, " showpage\n\n");
		fprintf (ps_file, "end\ngrestore\n");
	}
	else
		fprintf (ps_file, "\n");
	if (ps_file != stdout) fclose(ps_file);
}

/* fortran interface */
int ps_plotend_ (lastpage)
int *lastpage; {
	ps_plotend (*lastpage);
}

int ps_plotinit (plotfile, overlay, mode, xoff, yoff, xscl, yscl, ncopies, dpi, unit, page_width)
char *plotfile;		/* Name of output file or NULL for standard output */
double xoff, yoff;	/* Sets a new origin relative to old */
double xscl, yscl;	/* Global scaling, usually left to 1,1 */
double page_width;	/* Physical width of paper used (8.5 inch for regular laserplotter)
int overlay;		/* FALSE means print headers and macros first */
int mode;		/* 0 = Landscape, 1 = Portrait */
int ncopies;		/* Number of copies for this plot */
int dpi;		/* Plotter resolution in dots-per-inch */
int unit;		/* 0 = cm, 1 = inch, 2 = meter */
{
	int i;
	long int right_now, time();
	char openmode[2];
	double scl;

	ps_scl = (double)dpi;	/* Dots pr. inch resolution of output device */

	if (plotfile == NULL || plotfile[0] == 0)
		ps_file = stdout;
	else {
		(overlay) ? strcpy (openmode, "a") : strcpy (openmode, "w");
		if ((ps_file = fopen(plotfile, openmode)) == NULL) {
			fprintf (stderr, "pslib: Cannot create/open file : %s\n", plotfile);
			return (-1);
		}
	}
	right_now = time ((long *)0);
	fprintf (ps_file, "%%!PS-Adobe-2.0 EPSF-2.0\n");
	ps.mode = !(overlay && mode);
	ps.xscl = xscl;
	ps.xoff = xoff;
	ps.yscl = yscl;
	ps.yoff = yoff;
	
	if (!overlay) {
		/* Write definitions of macros to plotfile */
		if (mode == 0)
			fprintf (ps_file, "%%%%BoundingBox:0 0 792 612\n");
		else
			fprintf (ps_file, "%%%%BoundingBox:0 0 612 792\n");
		fprintf (ps_file, "%%%%Title: pslib v%lg document\n", Version);
		fprintf (ps_file, "%%%%Creator: Paul Wessel G&G/SOEST\n");
		fprintf (ps_file, "%%%%CreationDate: %s", ctime(&right_now));
		fprintf (ps_file, "%%%%EndComments\n\n");
		fprintf (ps_file, "gsave\n");
		
		fprintf (ps_file, "%% Begin header\n\n");
		
		fprintf (ps_file, "150 dict begin\n\n");
		fprintf (ps_file, "/A /setgray load def\n");
		fprintf (ps_file, "/B /setdash load def\n");
		fprintf (ps_file, "/C /setrgbcolor load def\n");
		fprintf (ps_file, "/D /rlineto load def\n");
		fprintf (ps_file, "/E {dup stringwidth pop} bind def\n");
		fprintf (ps_file, "/F /fill load def\n");
		fprintf (ps_file, "/G /rmoveto load def\n");
		fprintf (ps_file, "/L /lineto load def\n"); 
		fprintf (ps_file, "/M {stroke moveto} bind def\n");
		fprintf (ps_file, "/N /newpath load def\n");
		fprintf (ps_file, "/P /closepath load def\n");
		fprintf (ps_file, "/R /rotate load def\n");
		fprintf (ps_file, "/S /stroke load def\n");
		fprintf (ps_file, "/T /translate load def\n");
		fprintf (ps_file, "/U /grestore load def\n");
		fprintf (ps_file, "/V /gsave load def\n");
		fprintf (ps_file, "/W /setlinewidth load def\n");
		fprintf (ps_file, "/X {M dup 0 D dup -0.5 mul dup G 0 exch D S} bind def\n");
		fprintf (ps_file, "/Y {findfont exch scalefont setfont} bind def\n");
		fprintf (ps_file, "/Z /show load def\n");
		fprintf (ps_file, "/A0 {0 exch M 0 D D D D D 0 D P V A F U N} bind def\n");
		fprintf (ps_file, "/A1 {0 exch M 0 D D D D D 0 D P V A F U S} bind def\n");
		fprintf (ps_file, "/A2 {0 exch M 0 D D D D D 0 D P V C F U N} bind def\n");
		fprintf (ps_file, "/A3 {0 exch M 0 D D D D D 0 D P V C F U S} bind def\n");
		fprintf (ps_file, "/C0 {0 360 arc V A F U N} bind def\n");
		fprintf (ps_file, "/C1 {0 360 arc V A F U S} bind def\n");
		fprintf (ps_file, "/C2 {0 360 arc V C F U N} bind def\n");
		fprintf (ps_file, "/C3 {0 360 arc V C F U S} bind def\n");
		fprintf (ps_file, "/D0 {M 5 {dup} repeat D neg exch D neg exch neg D P V A F U N} bind def\n");
		fprintf (ps_file, "/D1 {M 5 {dup} repeat D neg exch D neg exch neg D P V A F U S} bind def\n");
		fprintf (ps_file, "/D2 {M 5 {dup} repeat D neg exch D neg exch neg D P V C F U N} bind def\n");
		fprintf (ps_file, "/D3 {M 5 {dup} repeat D neg exch D neg exch neg D P V C F U S} bind def\n");
		fprintf (ps_file, "/R0 {M dup 0 D exch 0 exch D neg 0 D P V A F U N} bind def\n");
		fprintf (ps_file, "/R1 {M dup 0 D exch 0 exch D neg 0 D P V A F U S} bind def\n");
		fprintf (ps_file, "/R2 {M dup 0 D exch 0 exch D neg 0 D P V C F U N} bind def\n");
		fprintf (ps_file, "/R3 {M dup 0 D exch 0 exch D neg 0 D P V C F U S} bind def\n");
		fprintf (ps_file, "/S0 {M dup dup 0 D 0 exch D neg 0 D P V A F U N} bind def\n");
		fprintf (ps_file, "/S1 {M dup dup 0 D 0 exch D neg 0 D P V A F U S} bind def\n");
		fprintf (ps_file, "/S2 {M dup dup 0 D 0 exch D neg 0 D P V C F U N} bind def\n");
		fprintf (ps_file, "/S3 {M dup dup 0 D 0 exch D neg 0 D P V C F U S} bind def\n");
		fprintf (ps_file, "/T0 {M dup 0 D dup -0.5 mul exch 0.866025 mul D P V A F U N} bind def\n");
		fprintf (ps_file, "/T1 {M dup 0 D dup -0.5 mul exch 0.866025 mul D P V A F U S} bind def\n");
		fprintf (ps_file, "/T2 {M dup 0 D dup -0.5 mul exch 0.866025 mul D P V C F U N} bind def\n");
		fprintf (ps_file, "/T3 {M dup 0 D dup -0.5 mul exch 0.866025 mul D P V C F U S} bind def\n");
		fprintf (ps_file, "/P0 {arc P V A F U N} bind def\n");
		fprintf (ps_file, "/P1 {arc P V A F U S} bind def\n");
		fprintf (ps_file, "/P2 {arc P V C F U N} bind def\n");
		fprintf (ps_file, "/P3 {arc P V C F U S} bind def\n");
		
		/* Define font macros (see pslib.h for details on how to add fonts) */
		
		for (i = 0; i < N_FONTS; i++) fprintf (ps_file, "/F%d {/%s Y} bind def\n", i, ps_font_name[i]);
		
		fprintf (ps_file, "/#copies %d def\n\n", ncopies);
		fprintf (ps_file, "currentdict\n\n");
		
		fprintf (ps_file, "%% Init coordinate system and scales\n");
		scl = 72.0 / ps_scl;
		fprintf (ps_file, "%% Scale is originally set to %lg, which means that\n", scl);
		if (unit == 0) {	/* CM used as unit */
			ps_scl /= 2.54;
			fprintf (ps_file, "%% 1 cm on the paper equals %d Postscript units\n", (int)ps_scl);
		}
		else if (unit == 1)	/* INCH used as unit */
			fprintf (ps_file, "%% 1 inch on the paper equals %d Postscript units\n", (int)ps_scl);
		else if (unit == 2) {	/* M used as unit */
			ps_scl /= 0.0254;
			fprintf (ps_file, "%% 1 m on the paper equals %d Postscript units\n", (int)ps_scl);
		}
		else {
			fprintf (stderr, "pslib: Measure unit not valid!\n");
			exit (-1);
		}
		
		xscl *= scl;
		yscl *= scl;
		if (mode == 0)	/* Landscape, 1 == Portrait */
			fprintf (ps_file, "%d 0 T 90 R\n", (int)rint (page_width * 72.0));
		fprintf (ps_file, "%lg %lg scale\n", xscl, yscl);
	}
	if (!(xoff == 0.0 && yoff == 0.0)) fprintf (ps_file, "%d %d T\n", (int)rint(xoff*ps_scl), (int)rint(yoff*ps_scl));
	ps_setpaint (0, 0, 0);
	fprintf (ps_file, "%% End of Header\n\n");
	ps_npath = 0;
	memset ((char *) ps_pattern_status, 0, 2 * N_PATTERNS);
	return (0);
}

/* fortran interface */
int ps_plotinit_ (plotfile, overlay, mode, xoff, yoff, xscl, yscl, ncopies, dpi, unit, page_width, nlen)
char *plotfile;	
double *xoff, *yoff;
double *xscl, *yscl;
double *page_width;
int *overlay;
int *mode;
int *ncopies;
int *dpi;
int *unit;
int nlen; {
	 ps_plotinit (plotfile, *overlay, *mode, *xoff, *yoff, *xscl, *yscl, *ncopies, *dpi, *unit, *page_width);
}

int ps_plotr (x, y, pen)
double x, y;
int pen; {
	int ix, iy;
	
	ix = rint (x * ps_scl);
	iy = rint (y * ps_scl);
	if (ix == 0 && iy == 0) return;
	ps_npath++;
	if (pen == 2)
		fprintf (ps_file, "%d %d D\n", ix, iy);
	else {
		fprintf (ps_file, "%d %d G\n", ix, iy);
		ps_npath = 1;
	}
}

/* fortran interface */
int ps_plotr_ (x, y, pen)
double *x, *y;
int *pen; {
	ps_plotr (*x, *y, *pen);
}

int ps_polygon (x, y, n, r, g, b, outline)
double *x, *y;
int n, r, g, b;
int outline; {
	double rr, gg, bb;
	
	rr = r * I_255;
	ps_line (x, y, n, 1, FALSE);	/* No stroke yet */
	ps_npath = 0;

	if (n > MAX_PATH) fprintf (stderr, "pslib: polygon called with > MAX_PATH points (%d)!\n", n);
	if (r == g && g == b && b == g)
		fprintf (ps_file, "P V %.3lg A F U", rr);
	else {
		gg = g * I_255;	bb = b * I_255;
		fprintf (ps_file, "P V %.3lg %.3lg %.3lg C F U", rr, gg, bb);
	}
	(outline) ? fprintf (ps_file, " S\n") : fprintf (ps_file, " N\n");
}

/* fortran interface */
int ps_polygon_ (x, y, n, r, g, b, outline)
double x[], y[];
int *n, *r, *g, *b;
int *outline; {
	ps_polygon (x, y, *n, *r, *g, *b, *outline);
}

int ps_rect (x1, y1, x2, y2, r, g, b, outline)
double x1, y1, x2, y2;
int r, g, b;
int outline; {
	int ix, iy, idx, idy;
	double rr, gg, bb;
	
	rr = r * I_255;
	ix = rint (x1 * ps_scl);
	iy = rint (y1 * ps_scl);
	idx = rint ((x2-x1) * ps_scl);
	idy = rint ((y2-y1) * ps_scl);
	if (r == g && g == b && b == g) /* Grayshade */
		fprintf (ps_file, "%.3lg %d %d %d %d R%d\n", rr, idy, idx, ix, iy, outline);
	else {
		gg = g * I_255;	bb = b * I_255;
		fprintf (ps_file, "%.3lg %.3lg %.3lg %d %d %d %d R%d\n", rr, gg, bb, idy, idx, ix, iy, outline+2);
	}
	ps_npath = 0;
}

/* fortran interface */
int ps_rect_ (x1, y1, x2, y2, r, g, b, outline)
double *x1, *y1, *x2, *y2;
int *r, *g, *b;
int *outline; {
	ps_rect (*x1, *y1, *x2, *y2, *r, *g, *b, *outline);
}

int ps_rotatetrans (x, y, angle)
double x, y, angle; {
	int ix, iy;
	int go = FALSE;
	
	ix = rint (x * ps_scl);
	iy = rint (y * ps_scl);
	if (angle != 0.0) {
		fprintf (ps_file, "%lg R", angle);
		go = TRUE;
	}
	if (ix != 0 || iy != 0) {
		if (go) putc (' ', ps_file);
		fprintf (ps_file, "%d %d T", ix, iy);
	}
	putc ('\n', ps_file);
}

/* fortran interface */
int ps_rotatetrans_ (x, y, angle)
double *x, *y, *angle; {
	 ps_rotatetrans (*x, *y, *angle);
}

int ps_setdash (pattern, offset)
char *pattern;	/* Line structure in Postscript units */
int offset; {	/* offset from plotpoint in " units */
	/* Examples:
	 * pattern = "4 4", offset = 0:
	 *   4 units of line, 4 units of space, start at current point
	 * pattern = "5 3 1 3", offset = 2:
	 *   5 units line, 3 units space, 1 unit line, 3 units space, start
	 *    2 units from curr. point.
	 */
	if (pattern)
		fprintf (ps_file, "S [%s] %d B\n", pattern, offset);
	else
		fprintf (ps_file, "S [] 0 B\n");	/* Reset to continous line */
	ps_npath = 0;
}

/* fortran interface */
int ps_setdash_ (pattern, offset, nlen)
char *pattern;
int nlen;
int *offset; {
	ps_setdash (pattern, *offset);
}

int ps_setfont (font_no)
int font_no; {
	if (font_no >= 0 && font_no < N_FONTS) ps.font_no = font_no;
}

/* fortran interface */
int ps_setfont_ (font_no)
int *font_no; {
	ps_setfont (*font_no);
}

int ps_setline (linewidth)
int linewidth; {
	fprintf (ps_file, "S %d W\n",linewidth);
}

/* fortan interface */
int ps_setline_ (linewidth)
int *linewidth; {
	 ps_setline (*linewidth);
}

int ps_setpaint (r, g, b)
int r, g, b; {
	fprintf (ps_file, "S %.3lg %.3lg %.3lg C\n", (r * I_255), (g * I_255), (b * I_255));
}

/* fortran interface */
int ps_setpaint_ (r, g, b)
int *r, *g, *b; {
	 ps_setpaint (*r, *g, *b);
}

int ps_square (x, y, side, r, g, b, outline)
double x, y, side;
int r, g, b;
int outline; {	/* see circle */
	int ds, ix, iy;
	double rr, gg, bb;
	
	rr = r * I_255;
	ix = rint ((x - 0.5*side) * ps_scl);
	iy = rint ((y - 0.5*side) * ps_scl);
	ds = rint (side*ps_scl);
	if (r == g && g == b && b == g) /* Grayshade */
		fprintf (ps_file, "%.3lg %d %d %d S%d\n", rr, ds, ix, iy, outline);
	else { /* Color */
		gg = g * I_255;	bb = b * I_255;
		fprintf (ps_file, "%.3lg %.3lg %.3lg %d %d %d S%d\n", rr, gg, bb, ds, ix, iy, outline+2);
	}
	ps_npath = 0;
}

/* fortran interface */
int ps_square_ (x, y, side, r, g, b, outline)
double *x, *y, *side;
int *r, *g, *b;
int *outline; {
	ps_square (*x, *y, *side, *r, *g, *b, *outline);
}

int ps_text (x, y, pointsize, text, angle, justify, form)
double x, y, angle;
char *text;
int pointsize;
int justify;	/* indicates what x,y refers to, see fig */
int form; {	/* 0 = normal text, 1 = outline */
  /*
   *    9	10	11
   *   |----------------|
   *   5  <textstring>  7
   *   |----------------|
   *   1	 2	 3
   */
	char label[512], tempstring[512], piece[512], *ptr;
	int dy, i, j, first, font;
	int sub, super, small;
	double height, small_size, size, scap_size, step, y0;

	if (strlen (text) >= 511) {
		fprintf (stderr, "pslib: text_item > 512 long!\n");
		return;
	}
	
	for (i = j = 0, first = TRUE; i < strlen (text); i++) {
		if (justify < 0 && text[i] == ' ' && first) continue;	/* Strip leading blanks */
		first = FALSE;
		if (text[i] == '(') {	/* Replace with \050 */
			label[j++] = '\\';
			label[j++] = '0';
			label[j++] = '5';
			label[j++] = '0';
		}
		else if (text[i] == ')') {	/* Replace with \051 */
			label[j++] = '\\';
			label[j++] = '0';
			label[j++] = '5';
			label[j++] = '1';
		}
		else
			label[j++] = text[i];
	}
	label[j] = 0;
	if (justify < 0) for (i = strlen (label) - 1; label[i] == ' '; i--) label[i] = 0;	/* Strip trailing blanks */
	
	justify = abs (justify);
  	height = pointsize / 72.0;
  	
  	/* Got to anchor point */
  	
  	ps_npath = 0;
	ps_ix = rint (x*ps_scl);
	ps_iy = rint (y*ps_scl);
	fprintf (ps_file, "%d %d M ", ps_ix, ps_iy);
	
	if (angle != 0.0) fprintf (ps_file, "V %.3lg R ", angle);
	y0 = -0.5 * height * ps_font_height[ps.font_no] * (justify/4);
	if (y0 != 0.0) fprintf (ps_file, "0 %d G ",(int) rint(y0 * ps_scl));
	if (!strchr (label, '@')) {	/* Plain text string */
		fprintf (ps_file, "%d F%d ", (int) rint (height * ps_scl), ps.font_no);
		fprintf (ps_file, "(%s) ", label);
		if ((justify %= 4) > 1) fprintf (ps_file, "E %d div 0 G ",(justify - 4));
		(form == 0) ? fprintf (ps_file, "Z") : fprintf (ps_file, "true charpath S");
		(angle != 0.0) ? fprintf (ps_file, " U\n") : fprintf (ps_file, "\n");
		return;
	}
	
	/* Here, we have special request for Symbol font and sub/superscript
	 * @~ toggles between Symbol font and default font
	 * @- toggles between subscript and normal text
	 * @+ toggles between superscript and normal text
	 * @# toggles between Small caps and normal text
	 */
	 
	font = ps.font_no;
	size = height;
	small_size = height * 0.7;
	scap_size = height * 0.85;
	step = 0.3 * height;
	sub = super = small = FALSE;
	
	if ((justify %= 4 ) > 1) {
		/* First we need to compute the justification.  The PS interpreter must do this for us */
		strcpy (tempstring, label);
		ptr = strtok (tempstring, "@");
		fprintf (ps_file, "0 ");	/* Initialize stringlength */
		while (ptr) {
			if (ptr[0] == '~') {	/* Symbol font toggle */
				font = (font == 12) ? ps.font_no : 12;
				ptr++;
				strcpy (piece, ptr);
			}
			else if (ptr[0] == '-') {	/* Subscript toggle  */
				sub = !sub;
				size = (sub) ? small_size : height;
				ptr++;
				strcpy (piece, ptr);
			}
			else if (ptr[0] == '+') {	/* Superscript toggle */
				super = !super;
				size = (super) ? small_size : height;
				ptr++;
				strcpy (piece, ptr);
			}
			else if (ptr[0] == '#') {	/* Small caps toggle */
				small = !small;
				size = (small) ? scap_size : height;
				ptr++;
				(small) ? get_uppercase (piece, ptr) : (int) strcpy (piece, ptr);
			}
			else
				strcpy (piece, ptr);
			if (strlen (piece) > 0) fprintf (ps_file, "%d F%d (%s) stringwidth pop add ",
				(int)rint(size*ps_scl), font, piece);
			ptr = strtok ((char *)NULL, "@");
		}
		/* Now move currentpoint to reflect the justification */
		fprintf (ps_file, "%d div 0 G\n", justify - 4);
	}
	
	/* Now we can start printing text items */
	
	sub = super = small = FALSE;
	ptr = strtok (label, "@");
	size = height;
	while (ptr) {
		if (ptr[0] == '~') {	/* Symbol font */
			font = (font == 12) ? ps.font_no : 12;
			ptr++;
			strcpy (piece, ptr);
		}
		else if (ptr[0] == '-') {	/* Subscript */
			sub = !sub;
			size = (sub) ? small_size : height;
			dy = (sub) ? rint(-step*ps_scl) : rint(step*ps_scl);
			fprintf (ps_file, "0 %d G\n", dy);
			ptr++;
			strcpy (piece, ptr);
		}
		else if (ptr[0] == '+') {	/* Superscript */
			super = !super;
			size = (super) ? small_size : height;
			dy = (super) ? rint(step*ps_scl) : rint(-step*ps_scl);
			fprintf (ps_file, "0 %d G\n", dy);
			ptr++;
			strcpy (piece, ptr);
		}
		else if (ptr[0] == '#') {	/* Small caps */
			small = !small;
			size = (small) ? scap_size : height;
			ptr++;
			(small) ? get_uppercase (piece, ptr) : (int) strcpy (piece, ptr);
		}
		else
			strcpy (piece, ptr);
		if (strlen (piece) > 0) {
			fprintf (ps_file, "%d F%d (%s)\n",(int)rint(size*ps_scl), font, piece);
			(form == 0) ? fprintf (ps_file, " Z\n") : fprintf (ps_file, " true charpath S\n");
		}
		ptr = strtok ((char *)NULL, "@");
	}
	if (angle != 0.0) fprintf (ps_file, "U\n");
}

/* fortan interface */
int ps_text_ (x, y, pointsize, text, angle, justify, form, nlen)
double *x, *y, *angle;
char *text;
int *pointsize;
int *justify;
int nlen;
int *form; {
	ps_text (*x, *y, *pointsize, text, *angle, *justify, *form);
}

int ps_textbox (x, y, pointsize, text, angle, justify, outline, dx, dy, r, g, b)
double x, y, angle;
char *text;
int pointsize, justify;	/* indicates what x,y refers to, see fig */
int outline;
double dx, dy;	/* Space between box border and text, in inches */
int r, g, b; {
  /*
   *    9	10	11
   *   |----------------|
   *   5  <textstring>  7
   *   |----------------|
   *   1	 2	 3
   */
	char label[512];
	int len, i, idx, idy, first = 0;
	double height, y0;

	if (strchr (text, '@')) return;	/* Text string contains escape sequence, not yet supported */
	
	if (strlen (text) >= 511) {
		fprintf (stderr, "pslib: text_item > 512 long!\n");
		return;
	}
	
	fprintf (ps_file, "\n%% ps_textbox begin:\n");
	
  	height = pointsize / 72.0;
	idx = rint (dx * ps_scl);	idy = rint (dy * ps_scl);
  	
  	/* Got to anchor point */
  	
	ps_ix = rint (x * ps_scl);
	ps_iy = rint (y * ps_scl);
	fprintf (ps_file, "V %d %d T 0 0 M ", ps_ix, ps_iy);
	
	strcpy (label, text);
	if (justify < 0) {	/* strip of leading and trailing blanks */
		len = strlen (label) - 1;
		for (i = len; i > 0 && label[i] == ' '; i--);
		label[++i] = 0;
		while (label[first] && label[first] == ' ') first++;
		justify = -justify;
	}
	y0 = -0.5 * height * ps_font_height[ps.font_no] * (justify/4);
	if (y0 != 0.0) fprintf (ps_file, "0 %d G ",(int)rint (y0 * ps_scl));
	fprintf (ps_file, "%d F%d (%s) ", (int)rint (height * ps_scl), ps.font_no, label);
	if ((justify %= 4) > 1) fprintf (ps_file, "E %d div 0 G ",(justify - 4));
	fprintf (ps_file, "false charpath flattenpath pathbbox\n");
	fprintf (ps_file, "%d add /y2 exch def %d add /x2 exch def %d sub /y1 exch def %d sub /x1 exch def N\n",
		idy, idx, idy, idx);
	if (angle != 0.0) fprintf (ps_file, "%.3lg R ", angle);
	fprintf (ps_file, "x1 y1 M x2 y1 L x2 y2 L x1 y2 L P V %.3lg %.3lg %.3lg C F U ", (r * I_255), (g * I_255), (b * I_255));
	(outline) ? fprintf (ps_file, "S U\n") : fprintf (ps_file, "N U\n");
	fprintf (ps_file, "%% ps_textbox end:\n\n");
}

/* fortran interface */
int ps_textbox_ (x, y, pointsize, text, angle, justify, outline, dx, dy, r, g, b, nlen)
double *x, *y, *angle;
char *text;
int *pointsize, *justify;
int *outline;
double *dx, *dy;
int nlen;
int *r, *g, *b; {
	 ps_textbox (*x, *y, *pointsize, text, *angle, *justify, *outline, *dx, *dy, *r, *g, *b);
}


int ps_transrotate (x, y, angle)
double x, y, angle; {
	int ix, iy;
	int go = FALSE;
	
	ix = rint (x * ps_scl);
	iy = rint (y * ps_scl);
	if (ix != 0 || iy != 0) {
		fprintf (ps_file, "%d %d T", ix, iy);
		go = TRUE;
	}
	if (angle != 0.0) {
		if (go) putc (' ', ps_file);
		fprintf (ps_file, "%lg R", angle);
	}
	putc ('\n', ps_file);
}

/* fortran interface */
int ps_transrotate_ (x, y, angle)
double *x, *y, *angle; {
	ps_transrotate (*x, *y, *angle);
}

int ps_triangle (x, y, side, r, g, b, outline)
double x, y, side;
int r, g, b;
int outline; {	/* see circle */
	int ix, iy, is;
	double rr, gg, bb;
	
	rr = r * I_255;
	ix = rint ((x-0.5*side) * ps_scl);
	iy = rint ((y-0.2886751*side) * ps_scl);
	is = rint (side * ps_scl);
	if (r == g && g == b && b == g) /* Grayshade */
		fprintf (ps_file, "%.3lg %d %d %d T%d\n", rr, is, ix, iy, outline);
	else { /* Color */
		gg = g * I_255;	bb = b * I_255;
		fprintf (ps_file, "%.3lg %.3lg %.3lg %d %d %d T%d\n", rr, gg, bb, is, ix, iy, outline+2);
	}
	ps_npath = 0;
}

/* fortran interface */
int ps_triangle_ (x, y, side, r, g, b, outline)
double *x, *y, *side;
int *r, *g, *b;
int *outline; {
	ps_triangle (*x, *y, *side, *r, *g, *b, *outline);
}

int ps_vector (xtail, ytail, xtip, ytip, tailwidth, headlength, headwidth, headshape, r, g, b, outline)
double xtail, ytail, xtip, ytip, tailwidth, headlength, headwidth, headshape;
int r, g, b;
int outline; {
	double angle, rr, gg, bb;
	int w2, length, hw, hl, hl2, hw2, l2;
	
	angle = atan2 ((ytip-ytail),(xtip-xtail)) * R2D;
	fprintf (ps_file, "V %d %d T", (int)rint (xtail * ps_scl), (int)rint (ytail * ps_scl));
	if (angle != 0.0) fprintf (ps_file, " %lg R", angle);
	rr = r * I_255;
	w2 = rint (0.5 * tailwidth * ps_scl);
	length = rint (hypot ((xtail-xtip), (ytail-ytip)) * ps_scl);
	hw = rint (headwidth * ps_scl);
	hl = rint (headlength * ps_scl);
	hl2 = rint (0.5 * headshape * headlength * ps_scl);
	hw2 = hw - w2;
	l2 = length - hl + hl2;
	if (r == g && g == b && b == g) /* grayshade */
		fprintf (ps_file, " %.3lg %d %d %d %d %d %d %d %d %d %d %d A%d U\n",
			rr, -l2, hl2, -hw2, -hl, hw, hl, hw, -hl2, -hw2, l2, -w2, outline);
	else {
		gg = g * I_255;	bb = b * I_255;
		fprintf (ps_file, " %.3lg %.3lg %.3lg %d %d %d %d %d %d %d %d %d %d %d A%d U\n",
			rr, gg, bb, -l2, hl2, -hw2, -hl, hw, hl, hw, -hl2, -hw2, l2, -w2, outline+2);
	}
}

/* fortran interface */
int ps_vector_ (xtail, ytail, xtip, ytip, tailwidth, headlength, headwidth, headshape, r, g, b, outline)
double *xtail, *ytail, *xtip, *ytip, *tailwidth, *headlength, *headwidth, *headshape;
int *r, *g, *b;
int *outline; {
	 ps_vector (*xtail, *ytail, *xtip, *ytip, *tailwidth, *headlength, *headwidth, *headshape, *r, *g, *b, *outline);
}

/* Support functions used in ps_* functions */

char *get_icon (file, nx, ny, invert)
char *file;
int *nx, *ny, invert; {
	int i, n_items, n_int, j, k, n_read, integer, n_alloc, n_lines, last;
	unsigned short int *pattern;
	char *out, t[8][7], line[80], width[12], height[12];
	FILE *fp;
	
	if ((fp = fopen (file, "r")) == NULL) {
		fprintf (stderr, "pslib: Cannot open file %s\n", file);
		return ((char *)NULL);
	}
	
	fgets (line, 80, fp);
	sscanf (&line[3], "%*[^,], %[^,], %[^,]", width, height);
	i = 0;	while (width[i] != '=') i++;
	*nx = atoi (&width[i+1]);
	i = 0;	while (height[i] != '=') i++;
	*ny = atoi (&height[i+1]);
	fgets (line, 80, fp);
	n_items = (*nx) * (*ny) / 4;
	n_int = n_items / 4;
	out = malloc ( (unsigned) (n_items+1));
	out[0] = 0;
	n_alloc = ceil (n_int / 8.0) * n_int;
	pattern = (unsigned short int *) malloc ((unsigned)n_alloc);
	n_lines = n_items / 32;
	for (i = k = 0; i < n_lines; i++) {
		fgets (line, 80, fp);
		last = (i == (n_lines - 1));
		if (last) {
			n_read = sscanf (&line[1], "%[^,],%[^,],%[^,],%[^,],%[^,],%[^,],%[^,],%[^,]", t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7]);
			if (n_read == 4) sscanf (&line[1], "%[^,],%[^,],%[^,],%[^,]", t[4], t[5], t[6], t[7]);	/* New style icon file */
		}
		else {
			n_read = sscanf (&line[1], "%[^,],%[^,],%[^,],%[^,],%[^,],%[^,],%[^,],%[^,],", t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7]);
			if (n_read == 4) sscanf (&line[1], "%[^,],%[^,],%[^,],%[^,],", t[4], t[5], t[6], t[7]);	/* New style icon file */
		}
		for (j = 0; j < 8; j++, k++) {
			sscanf (&t[j][2], "%x", &integer);
			pattern[k] = integer;
		}
	}
	fclose (fp);
	if (invert) for (k = 0; k < n_int; k++) pattern[k] = ~pattern[k];
	for (k = 0; k < n_int; k++) {
		sprintf (t[0], "%.4x", (int)pattern[k]);
		strcat (out, t[0]);
	}
	out[n_items] = 0;
	free ((char *)pattern);
	return (out);
}

int get_uppercase (new, old)
char *new, *old; {
	int i = 0;
	while (old[i]) {
		new[i] = toupper ((int)old[i]);
		i++;
	}
	new[i] = 0;
}
