C*****************************************************************************
C*****                                                                   *****
C*****                             RAND.FTN                              *****
C*****                                                                   *****
C*****               Pseudo Random Number Generator Package              *****
C*****                            Version 3                              *****
C*****               David M. Krowitz August 25, 1986.                   *****
C*****                                                                   *****
C*****      Copyright (c) 1986                                           *****
C*****      David M. Krowitz                                             *****
C*****      Massachusetts Institute of Technology                        *****
C*****      Department of Earth, Atmosheric, and Planetary Sciences      *****
C*****************************************************************************   
C****************************************************************************
C
C           Routine to seed the random number generator using
C           the system's time of day. 
C                         
C****************************************************************************
c
c      SUBROUTINE RAND_$INIT
c
c      INTEGER*4 I,J,K
c      INTEGER*2 CLOCK(3)
c
c      COMMON /RANDCOM/ INITFLAG,RANTAB,INDEX1,INDEX2,INDEX3
c      LOGICAL INITFLAG
c      INTEGER*4 RANTAB(0:54)
c      INTEGER*4 INDEX1,INDEX2,INDEX3
c      DATA INITFLAG /.FALSE./
c
C
C   Get the 48-bit long time of day from the system and turn it
C   into a 32-bit long integer.
C
c      CALL TIME_$CLOCK (CLOCK)
c      I = INT4(CLOCK(1))
c      J = INT4(CLOCK(2))*2**4
c      K = INT4(CLOCK(3))*2**8
c      RANTAB(0) = I+J+K
C
C   Use a linear congruential psuedo random number generator to
C   initialize the table for the RAND_$UNIF function.
C
c      DO 100 I = 1,54
c        RANTAB(I) = RANTAB(I-1)*31415821+1
c100   CONTINUE
c
c     INITFLAG = .TRUE.
c      INDEX1 = 0
c      INDEX2 = 23
c      INDEX3 = 54
c
c      RETURN
c      END
c
C****************************************************************************
C
C           Uniform Random Number Generator.
C           Generates a psuedo-random number in the range 0.0 to 1.0 with
C           a uniform distribution. The first time RAND_$UNIF is called the
C           argument is used to seed the random number generator if the
C           random number generator has not already been seeded by a call
C           to the RAND_$INIT routine. If the seed passed on the first call
C           is greater than 0, then the seed is used to initialize the
C           random number generator, otherwise a built in seed number is
C           used to set up the initial table. Since IEEE standard floating
C           point numbers have a 24 bit mantissa (one bit being the hidden bit)
C           we generate random integers in the range 0 to 2**24-1 and then
C           divide them by 2**24-1 to get a real number in the range 0 to
C           1.0, inclusive.
C
C****************************************************************************

      REAL*4 FUNCTION RAND_UNIF (ISEED)

      COMMON /RANDCOM/ INITFLAG,RANTAB,INDEX1,INDEX2,INDEX3
      LOGICAL INITFLAG
      INTEGER*4 RANTAB(0:54)
      INTEGER*4 INDEX1,INDEX2,INDEX3

      INTEGER*4 ISEED

C
C   If random number generator has not already been seeded
C   by calling the RAND_$INIT routine, then set up the initial
C   table for the additive congruential method used here. If
C   the seed number supplied by the caller is 0, then use our
C   own seed number.
C
      IF (INITFLAG) GOTO 100

      IF (ISEED.GT.0) THEN
        RANTAB(0) = ISEED
      ELSE
        RANTAB(0) = 31415926
      ENDIF

      DO 10 I = 1,54
        RANTAB(I) = RANTAB(I-1)*31415821+1
10    CONTINUE

      INITFLAG = .TRUE.
      INDEX1 = 0
      INDEX2 = 23
      INDEX3 = 54

C
C   Generate the next random integer in the range 0 to 2**24-1
C   and then convert it into a real number in the range 0.0 to 1.0
C
100   INDEX1 = MOD(INDEX1+1,55)
      INDEX2 = MOD(INDEX2+1,55)
      INDEX3 = MOD(INDEX3+1,55)
      RANTAB(INDEX1) = RANTAB(INDEX2)+RANTAB(INDEX3)

      RAND_UNIF = FLOAT(IAND(RANTAB(INDEX1),2**24-1))/FLOAT(2**24-1)

      RETURN
      END




C****************************************************************************
C
C           Vector Uniform Random Number Generator.
C           Same as RAND_$UNIF function except it fills a complete table
C           full of uniformly distributed pseudo-random numbers in the
C           range 0.0 to 1.0. This routine avoids the overhead of making
C           repeated functions calls when you need a lot of random numbers
C           at once.
C
C****************************************************************************

      SUBROUTINE RAND_VEC_UNIF (ISEED,TABLE,NUM)

      COMMON /RANDCOM/ INITFLAG,RANTAB,INDEX1,INDEX2,INDEX3
      LOGICAL INITFLAG
      INTEGER*4 RANTAB(0:54)
      INTEGER*4 INDEX1,INDEX2,INDEX3

      INTEGER*4 ISEED,NUM
      REAL*4 TABLE(NUM)

C
C   If random number generator has not already been seeded
C   by calling the RAND_$INIT routine, then set up the initial
C   table for the additive congruential method used here. If
C   the seed number supplied by the caller is 0, then use our
C   own seed number.
C
      IF (INITFLAG) GOTO 100

      IF (ISEED.GT.0) THEN
        RANTAB(0) = ISEED
      ELSE
        RANTAB(0) = 31415926
      ENDIF

      DO 10 I = 1,54
        RANTAB(I) = RANTAB(I-1)*31415821+1
10    CONTINUE

      INITFLAG = .TRUE.
      INDEX1 = 0
      INDEX2 = 23
      INDEX3 = 54

C
C   Generate the next random integer in the range 0 to 2**24-1
C   and then convert it into a real number in the range 0.0 to 1.0
C
100   DO 200 I = 1,NUM
        INDEX1 = MOD(INDEX1+1,55)
        INDEX2 = MOD(INDEX2+1,55)
        INDEX3 = MOD(INDEX3+1,55)
        RANTAB(INDEX1) = RANTAB(INDEX2)+RANTAB(INDEX3)
        TABLE(I) = FLOAT(IAND(RANTAB(INDEX1),2**24-1))/FLOAT(2**24-1)
200   CONTINUE

      RETURN
      END


C****************************************************************************
C
C           Gaussian Random Number Generator.
C           Generate random numbers in the range -1.0E+38 to 1.0E+38 (largest
C           negative single precision real number to largest positive single
C           precision real number) with a distribution of the form:
C
C                                              2
C                  1             -1  ( X - MU )
C           _______________      --- (--------)     , SIGMA > 0 == STD. DEVIATION
C              _____              2  ( SIGMA  )
C             /                E                    , MU == MEAN VALUE
C           \/ 2 PI   SIGMA
C                         
c  third moment perterbation to Gaussian calculated using hermite polynomials
c
C****************************************************************************
                          
      REAL FUNCTION RAND_GAUSSIAN (SIGMA,MU)
                          
      REAL PI
      REAL DELTA                      
      INTEGER SAMPLES
      REAL SIGMA,MU
      PARAMETER (PI = 3.1415926)
      PARAMETER (DELTA = 6.0E0)
      PARAMETER (SAMPLES = 200)

      LOGICAL LINIT
      INTEGER*2 INTV(SAMPLES)
      REAL RAND_UNIF     
      DOUBLE PRECISION X(SAMPLES),ERF(SAMPLES)
      DOUBLE PRECISION IERF,DX
      SAVE LINIT           
      SAVE INTV
      SAVE X,ERF   

      DATA LINIT/.FALSE./ 
C               
C           Define the unit Gaussian function.
C                           
      GAUSS(X) = DEXP(-1.0E0/2.0E0*X**2)/(SQRT(2.0E0*PI))

C
C           If not already initialized, tabulate the integral of the
C           unit gaussian funtion using Simpson's method.
C
      IF (LINIT) GOTO 1000
      DX = 2.0E0*DELTA/FLOAT(SAMPLES-1)
      X(1) = -DELTA
      ERF(1) = 0.0E0
      DO 10 I = 2,SAMPLES
        X(I) = X(I-1)+DX
        ERF(I) = ERF(I-1)+(GAUSS(X(I-1))+4.0E0*
     &            GAUSS(X(I)-DX/2.0E0)+GAUSS(X(I)))*DX/6.0E0
10    CONTINUE

C
C           Set up a table for looking up what interval of
C           the inverse ERF function that a particular value
C           would fall into (since the ERF function was
C           evaluated at equally spaced intervals, its
C           inverse does not have equally space points).
C

      INTERVAL =1
      DO 30 I = 1,SAMPLES
        IERF = (I-1)*(1.0E0/FLOAT(SAMPLES-1))
        if (interval .ge. samples) goto 20 
15      IF (IERF.LT.ERF(INTERVAL)) GOTO 20
        INTERVAL = INTERVAL+1
        if (interval .ge. samples) goto 20 
        GOTO 15
20      INTV(I) = INTERVAL
30    CONTINUE

C
C           Initialization done.
C
      LINIT = .TRUE.
    
C
C           Look up the inverse ERF funtion for a given random number
C           between 0.0 and 1.0 and interpolate result.
C

1000  R = RAND_UNIF(0)
      INTERVAL = INTV(short(R*(SAMPLES-1)+1))
      if (interval .ge. samples) then
         interval = samples
         goto 1100
      endif   
      if (interval .lt. 1) interval = 1
1010  IF (R.LT.ERF(INTERVAL)) GOTO 1100
      INTERVAL = INTERVAL+1
      if (interval .ge. samples) then
         interval = samples
         goto 1100
      endif
      GOTO 1010

1100  XX = X(INTERVAL-1)+(X(INTERVAL)-X(INTERVAL-1))/
     &     (ERF(INTERVAL)-ERF(INTERVAL-1))*(R-ERF(INTERVAL-1))

      RAND_GAUSSIAN = SIGMA*XX+MU

      RETURN 
      END 


C****************************************************************************
C
C           Vector Gaussian Random Number Generator.
C           Same as the RAND_$GAUSSIAN function except that it fills a
C           complete table of pseudo-random numbers with a distribution
C           of a Gaussian function. This routine avoids the overhead of
C           making repeated function calls when you need a lot of values
C           all at once.
C                         
C****************************************************************************
                          
      SUBROUTINE RAND_VEC_GAUSSIAN (SIGMA,MU,TABLE,NUM)
      
      parameter (ndmax=100000)         
      REAL PI
      REAL DELTA                      
      INTEGER SAMPLES
      PARAMETER (PI = 3.1415926)
      PARAMETER (DELTA = 6.0E0)
      PARAMETER (SAMPLES = 200)

      REAL SIGMA,MU
      INTEGER*4 NUM
      REAL TABLE(5*ndmax)

      LOGICAL INIT        
      INTEGER*2 INTV(SAMPLES)
      REAL RAND_UNIF     
      DOUBLE PRECISION X(SAMPLES),ERF(SAMPLES)
      DOUBLE PRECISION IERF,DX
      SAVE INIT           
      SAVE INTV
      SAVE X,ERF

      DATA INIT/.FALSE./
C
C           Define the unit Gaussian function.
C
      GAUSS(X) = 1.0E0/(SQRT(2.0E0*PI))*DEXP(-1.0E0/2.0E0*X**2)

C
C           If not already initialized, tabulate the integral of the
C           unit gaussian funtion using Simpson's method.
C
      IF (INIT) GOTO 1000
      DX = 2.0E0*DELTA/FLOAT(SAMPLES-1)
      X(1) = -DELTA
      ERF(1) = 0.0E0
      DO 10 I = 2,SAMPLES
        X(I) = X(I-1)+DX
        ERF(I) = ERF(I-1)+(GAUSS(X(I-1))+4.0E0*GAUSS(X(I)-DX/2.0E0)
     &                     +GAUSS(X(I)))*DX/6.0E0
10    CONTINUE

C
C           Set up a table for looking up what interval of
C           the inverse ERF function that a particular value
C           would fall into (since the ERF function was
C           evaluated at equally spaced intervals, its
C           inverse does not have equally space points).
C

      INTERVAL =1
      DO 30 I = 1,SAMPLES
        IERF = (I-1)*(1.0E0/FLOAT(SAMPLES-1)) 
        if (interval .ge. samples) goto 20 
15      IF (IERF.LT.ERF(INTERVAL)) GOTO 20
        INTERVAL = INTERVAL+1
        if (interval .ge. samples) goto 20 
        GOTO 15
20      INTV(I) = INTERVAL
30    CONTINUE

C
C           Initialization done.
C
      INIT = .TRUE.
    
C
C           Look up the inverse ERF funtion for a given random number
C           between 0.0 and 1.0 and interpolate result.
C

1000  DO 2000 I = 1,NUM
        R = RAND_UNIF(0)
        INTERVAL = INTV(short(R*(SAMPLES-1)+1))
        if (interval .ge. samples) then
           interval = samples
           goto 1100
        endif   
1010    IF (R.LT.ERF(INTERVAL)) GOTO 1100
        INTERVAL = INTERVAL+1 
        if (interval .ge. samples) then
           interval = samples
           goto 1100
        endif
        GOTO 1010

1100    XX = X(INTERVAL-1)+(X(INTERVAL)-X(INTERVAL-1))/
     &       (ERF(INTERVAL)-ERF(INTERVAL-1))*(R-ERF(INTERVAL-1))

        TABLE(I) = SIGMA*XX+MU
2000  CONTINUE

      RETURN
      END  

C****************************************************************************
C
C           Hermite Polynomial Random Number Generator.
C           Generate random numbers in the range -1.0E+38 to 1.0E+38 (largest
C           negative single precision real number to largest positive single
C           precision real number) with a distribution of the form:
C
C                                              2
C                  1             -1  ( X - MU )
C           _______________      --- (--------)     , SIGMA > 0 == STD. DEVIATION
C              _____              2  ( SIGMA  )
C             /                E                    , MU == MEAN VALUE
C           \/ 2 PI   SIGMA
C                             
c  3rd and 4th moment perterbation to Gaussian calculated using hermite polynomials
c  sigma and mu are assumed to be 1 and 0, but may be easily changed.
c  reinit must be set to .true. any time this function is called with different
c  mu3 and mu4 from the last call.
c
C****************************************************************************
                          
      REAL FUNCTION RAND_HERMITE (mu3,mu4,reinit)
                          
      REAL PI
      REAL DELTA                      
      INTEGER SAMPLES
      REAL SIGMA,MU,mu3,mu4
      PARAMETER (PI = 3.1415926)
      PARAMETER (DELTA = 6.0E0)
      PARAMETER (SAMPLES = 2000)

      LOGICAL LINIT,reinit        
      INTEGER*2 iINTV(SAMPLES)
      REAL RAND_UNIF     
      real xX(SAMPLES),xERF(SAMPLES)
      real IERF,DX,xmu3,xmu4
      SAVE LINIT           
      SAVE iINTV
      SAVE xX,xERF   

      DATA LINIT/.FALSE./ 
      if (reinit) then
         linit = .false.
      endif   
      sigma = 1.0
      mu = 0.0
      xmu3 = mu3
      xmu4 = mu4
C
C           If not already initialized, tabulate the integral of the
C           unit gaussian funtion using Simpson's method.
C
      IF (LINIT) GOTO 1000
      DX = 2.0E0*DELTA/FLOAT(SAMPLES-1)
      xX(1) = -DELTA
      xERF(1) = 0.0E0
      DO 10 I = 2,SAMPLES
        xX(I) = xX(I-1)+DX
        xERF(I) = xERF(I-1)+(HERMITE(xX(I-1),xmu3,xmu4)+4.0E0*
     &            HERMITE(xX(I)-DX/2.0E0,xmu3,xmu4) + 
     &            HERMITE(xX(I),xmu3,xmu4))*DX/6.0E0
c        xerf(i) = xerf(i-1) + HERMITE(xx(i),xmu3,xmu4)*dx
c        if (xerf(i) .lt. 0.) then
c           xerf(i) = 0.
c        endif
10    CONTINUE

C
C           Set up a table for looking up what interval of
C           the inverse ERF function that a particular value
C           would fall into (since the ERF function was
C           evaluated at equally spaced intervals, its
C           inverse does not have equally space points).
C

      INTERVAL =1
      DO 30 I = 1,SAMPLES
        IERF = (I-1)*(1.0E0/FLOAT(SAMPLES-1))  
        if (interval .ge. samples) goto 20 
15      IF (IERF.LT.xERF(INTERVAL)) GOTO 20
        INTERVAL = INTERVAL+1
        if (interval .ge. samples) goto 20 
        GOTO 15
20      iINTV(I) = INTERVAL
30    CONTINUE

C
C           Initialization done.
C
      LINIT = .TRUE.
    
C
C           Look up the inverse ERF funtion for a given random number
C           between 0.0 and 1.0 and interpolate result.
C

1000  R = RAND_UNIF(0)
      INTERVAL = iINTV(short(R*(SAMPLES-1)+1)) 
      if (interval .ge. samples) then
         interval = samples
         goto 1100
      endif   
      if (interval .lt. 1) interval = 1
1010  IF (R.LT.xERF(INTERVAL)) GOTO 1100
      INTERVAL = INTERVAL+1  
      if (interval .ge. samples) then
         interval = samples
         goto 1100
      endif
      GOTO 1010

1100  xXX = xX(INTERVAL-1)+(xX(INTERVAL)-xX(INTERVAL-1))/
     &     (xERF(INTERVAL)-xERF(INTERVAL-1))*(R-xERF(INTERVAL-1))

      RAND_HERMITE = sigma*xXX + mu

      RETURN 
      END 






